import math
import numpy as np
import torch

from tqdm import tqdm
from treelib import Tree

device = "cuda" if torch.cuda.is_available() else "cpu"


class Node:
    def __init__(self, state=None, name="", index=0):
        self.value = 0
        self.n = 0
        self.state = state
        self.name = name
        self.index = index
        self.children = []

    def get_max_ucb1_child(self):
        if not self.children:
            return None

        max_node = self.children[0]
        max_ucb1 = float("-inf")

        for child in self.children:
            ucb1 = self.get_child_ucb1(child)

            if ucb1 > max_ucb1:
                max_ucb1 = ucb1
                max_node = child

        return max_node

    def get_child_ucb1(self, child):
        if child.n == 0:
            return float("inf")
        return child.value / child.n + 2 * math.sqrt(math.log(self.n, math.e) / child.n)

    def display_tree(self, stdout=True):
        tree = Tree()

        tree.create_node(f"Root => value: {self.value}, n: {self.n}", "root")
        stack = [(self, "root")]
        count = 0

        while stack:
            cur_node, node_name = stack.pop()
            for child in cur_node.children:
                tree.create_node(
                    f"{child.name} => value: {child.value}, n: {child.n}", f"node_{count}", parent=node_name
                )
                stack.append((child, f"node_{count}"))
                count += 1

        return tree.show(stdout=stdout)


class MCTS:
    def __init__(self, env, agent, image, mask, reset=False):
        self.env = env
        self.agent = agent

        if reset:
            start_state = self.env.reset(image, mask)
        else:
            start_state = self.env.get_cur_state()
        self.start_env = self.env.copy()
        self.root_node = Node(start_state)

        for act in self.env.get_valid_actions():
            env_copy = self.env.copy()
            new_state, _, _ = env_copy.step(act)
            new_node = Node(new_state, self.env.actions[act], act)
            self.root_node.children.append(new_node)

    def run(self, n_iter=200):
        for _ in tqdm(range(n_iter)):
            value, node_path = self.traverse()
            self.backpropagate(node_path, value)
            self.env = self.start_env.copy()

        # self.root_node.display_tree()
        vals = [float("-inf")] * len(self.env.actions)
        for child in self.root_node.children:
            vals[child.index] = (child.value / child.n) if child.n else 0
        exp_vals = np.exp((np.array(vals) + 1) * 3)  # Scale to amplify difference
        return exp_vals / sum(exp_vals)

    def traverse(self):
        cur_node = self.root_node
        node_path = [cur_node]
        while cur_node.children:
            cur_node = cur_node.get_max_ucb1_child()
            self.env.step(cur_node.index)
            node_path.append(cur_node)

        if cur_node.n:
            for act in self.env.get_valid_actions():
                env_copy = self.env.copy()
                new_state, _, _ = env_copy.step(act)
                new_node = Node(new_state, self.env.actions[act], act)
                cur_node.children.append(new_node)
            while cur_node.children:
                cur_node = cur_node.get_max_ucb1_child()
                node_path.append(cur_node)
        # print("Node path:", [node.name for node in node_path])

        return self.rollout(cur_node), node_path

    def backpropagate(self, node_path: list, last_value: float):
        for node in node_path[::-1]:
            node.value += last_value
            node.n += 1

    def rollout(self, state_node: Node) -> float:
        if state_node.name == "=":
            return self.env.get_reward(done=True)

        tot_reward = 0
        cur_state = state_node.state

        step = 0
        while True:
            # print(step, tot_reward, self.env.cur_exp)
            step += 1
            state = torch.tensor(np.expand_dims(cur_state, 0)).int().to(device)
            with torch.no_grad():
                probs = self.agent.generate_single(state).squeeze()

            invalid_acts = self.env.get_invalid_actions()
            probs[invalid_acts] = -1

            action_idx = probs.argmax()

            cur_state, reward, done = self.env.step(action_idx)
            tot_reward += reward

            if done:
                break

        return tot_reward



import numpy as np
import random


def eval_expression(exp: list, image: np.ndarray = None):
    expression = ""

    for token in exp:
        if token[0] == "c":
            channel = eval(token[1:])
            expression += f"(image[{channel}] + {(random.random() * 0.001 + 1) * 1e-5})"  # To prevent divide by zero
        elif token == "sq":
            expression += "**2"
        elif token == "sqrt":
            expression += "**0.5"
        elif token == "=":
            break
        else:
            expression += token

    try:
        return eval(expression)
    except (SyntaxError, FloatingPointError):
        return False
    except TypeError:
        return False


class UnitNode(object):
    def __init__(self, level=0):
        self.level = level

    def __add__(self, obj):
        if obj.level == self.level:
            return UnitNode(self.level)
        raise ValueError

    def __sub__(self, obj):
        return self + obj

    def __truediv__(self, obj):
        return UnitNode(self.level - obj.level)

    def __mul__(self, obj):
        return UnitNode(self.level + obj.level)

    def __pow__(self, val):
        return UnitNode(self.level * val)


def check_unitless_validity(exp: list):
    expression = ""
    for token in exp:
        if token[0] == "c":
            expression += "UnitNode(1)"
        elif token == "sq":
            expression += "**2"
        elif token == "sqrt":
            expression += "**0.5"
        elif token == "-":
            expression += "+"
        elif token == "=":
            break
        else:
            expression += token
    try:
        res = eval(expression)
        if res.level == 0:
            return True
        return res.level
    except ValueError:
        return 1
    except (SyntaxError, FloatingPointError):
        return False


if __name__ == "__main__":
    # image = np.random.rand(3, 2, 2)
    # print(image)
    # print(eval_expression(["(", "c0", "/", "c2", ")"], image))
    print(check_unitless_validity(["(", "c0", "+", "c1", ")", "/", "(", "c2", "+", "c2", ")"]))



from glob import glob
import os
import gym
from copy import deepcopy
import numpy as np
from tqdm import tqdm
import pickle

from sklearn.metrics import roc_curve, auc

from indexrl.expression_handler import check_unitless_validity, eval_expression
from indexrl.utils import standardize


class IndexRLEnv(gym.Env):
    def __init__(
        self,
        discrete_actions: list,
        max_exp_len: int = 100,
        ohe: bool = True,
        masked_actions: list = None,
        unitless: bool = False,
    ):
        super(IndexRLEnv, self).__init__()
        self.actions = discrete_actions
        self.image = None
        self.mask = None
        self.cur_exp = []
        self.parentheses_level = 0
        self.max_exp_len = max_exp_len
        self.ohe = ohe
        self.masked_actions = masked_actions
        self.unitless = unitless

        self.best_reward = 0
        self.best_exp = []
        self.seen = set()

    def load_seen(self, seen_path):
        with open(seen_path, "rb") as fp:
            self.seen = pickle.load(fp)

    def save_seen(self, seen_path):
        print("Seen length:", len(self.seen))
        with open(seen_path, "wb") as fp:
            pickle.dump(self.seen, fp)

    def get_cur_state(self):
        if self.ohe:
            cur_exp_indices = [self.actions.index(act) for act in self.cur_exp] + [
                0
            ] * (self.max_exp_len - len(self.cur_exp))
            enc_state = np.zeros((self.max_exp_len, len(self.actions)))
            enc_state[np.arange(self.max_exp_len), cur_exp_indices] = 1
            return enc_state.flatten()
        else:
            cur_exp_indices = [self.actions.index(act) for act in self.cur_exp]
            return np.array(cur_exp_indices)

    def step(self, action_idx: int) -> tuple:
        """Take a step in the environment with the specified action.

        Args:
            action_idx (int): discrete action index

        Returns:
            np.ndarray: current state
            float:      reward
            bool:       done
        """
        done = self.take_action(action_idx)

        if len(self.cur_exp) >= self.max_exp_len:
            done = True

        if done:
            exp_s = str(self.cur_exp)
            if exp_s in self.seen:
                return self.get_cur_state(), -1, True
            self.seen.add(exp_s)

        reward = self.get_reward(done)

        return self.get_cur_state(), reward, done

    def reset(self, image: np.ndarray = None, mask: np.ndarray = None) -> np.ndarray:
        if image is not None and mask is not None:
            self.image = image
            self.mask = mask
        self.cur_exp = []
        self.parentheses_level = 0

        return self.get_cur_state()

    def render(self):
        print(self.cur_exp)

    def get_reward(self, done: bool) -> float:
        if not done:
            return 0

        if len(self.cur_exp) < 3:
            return -1

        unitless = check_unitless_validity(self.cur_exp)
        result = eval_expression(self.cur_exp, self.image.squeeze())
        if result is False:
            return -1

        if len(self.mask) > 2:
            reward = get_auc_f1(result, self.mask > 0)
            # reward = (np.abs(result - self.mask) < 0.1).sum() / len(self.mask)
            self.best_reward = max(self.best_reward, reward)
            self.best_exp = self.cur_exp
            return reward
        else:  # Pretraining stage
            # Motivate longer expressions
            if len(self.cur_exp) < self.max_exp_len // 3 or "(" not in self.cur_exp:
                return -1
            return (
                0.5
                + 0.05 * len(self.cur_exp)
                + 0.01 * self.cur_exp.count(")")
                + 1 * unitless
            )

    def take_action(self, action_idx: int) -> bool:
        action = self.actions[action_idx]
        if action == "(":
            self.parentheses_level += 1
        elif action == ")":
            self.parentheses_level -= 1
        self.cur_exp.append(action)

        return action == "="

    def get_valid_actions(self):
        if len(self.cur_exp) == self.max_exp_len - 1:
            return {self.actions.index("=")}

        # Include all the channels and opening brackets in action set 1
        acts_1 = []
        for i, act in enumerate(self.actions):
            if act[0] == "c" or act in ("("):
                acts_1.append(i)

        # Remove actions specified as masked actions from acts_1
        if self.masked_actions:
            for act in self.masked_actions:
                idx = self.actions.index(act)
                if idx in acts_1:
                    acts_1.remove(idx)

        # Include the inverse of the action set 1 in action set 2
        acts_2 = list(set(range(len(self.actions))) - set(acts_1))

        # Allow action set 1 when just starting the episode
        if not self.cur_exp:
            return acts_1

        last_act = self.cur_exp[-1]

        # Disallow any other actions if episode ended with "="
        if last_act == "=":
            return []

        # If last action was one of the following, allow selecting channels or an open parenthesis
        if last_act in list("(+-*/"):
            return acts_1

        # Disallow consecutive squares, square roots, or combinations of them.
        if last_act == "sq" or last_act == "sqrt":
            acts_2.remove(self.actions.index("sq"))
            acts_2.remove(self.actions.index("sqrt"))

        # Remove actions specified as masked actions from acts_2
        if self.masked_actions:
            for act in self.masked_actions:
                idx = self.actions.index(act)
                if idx in acts_2:
                    acts_2.remove(idx)

        # Disallow closing the brackets after just one character
        if len(self.cur_exp) > 1 and self.cur_exp[-2] == "(":
            acts_2.remove(self.actions.index(")"))

        # Disallow closing paranthesis if there are no open paranthesis
        if self.parentheses_level <= 0:
            acts_2.remove(self.actions.index(")"))
        else:
            acts_2.remove(self.actions.index("="))

        return acts_2

    def get_invalid_actions(self):
        return list(set(range(len(self.actions))) - set(self.get_valid_actions()))

    def copy(self):
        return deepcopy(self)

    def state_to_expression(self, state):
        return list(map(lambda x: self.actions[x], state))


def get_final_reward(exp: list, image_dir: str, mask_dir: str, rew_type="-") -> float:
    image_paths = glob(os.path.join(image_dir, "*.npy"))
    mask_paths = glob(os.path.join(mask_dir, "*.npy"))
    assert len(image_paths) == len(mask_paths)

    unitless = check_unitless_validity(exp)
    if unitless is False:
        return -1

    tot_reward = 0
    for image_path, mask_path in tqdm(
        zip(image_paths, mask_paths), "Calculating reward"
    ):
        image = np.load(image_path)
        mask = np.load(mask_path)

        result = eval_expression(exp, image.squeeze())
        if result is False:
            tot_reward -= 1
        elif len(mask) > 2:
            norm_result = standardize(result, 3, (0, 1))
            if rew_type == "corr":
                tot_reward += abs(get_correlation(1 - norm_result, mask > 0))
            elif rew_type == "f1":
                tot_reward += min(
                    get_f1_score(norm_result, mask > 0),
                    get_f1_score(1 - norm_result, mask > 0),
                )
            elif rew_type == "auc":
                tot_reward += min(
                    get_auc_score(norm_result, mask > 0),
                    get_auc_score(1 - norm_result, mask > 0),
                )
            elif rew_type == "sim":
                tot_reward += min(
                    get_similarity(norm_result, mask > 0),
                    get_similarity(1 - norm_result, mask > 0),
                )
            elif rew_type == "iou":
                tot_reward += min(
                    get_jaccard(norm_result, mask > 0),
                    get_jaccard(1 - norm_result, mask > 0),
                )
            else:  # AUC F1
                tot_reward += min(
                    get_auc_f1(standardize(result), mask > 0),
                    get_auc_f1(1 - standardize(result), mask > 0),
                )
        else:  # Pretraining stage
            tot_reward += 0.5 + 0.02 * len(exp) + 0.2 * exp.count(")") + 1 * unitless

    return tot_reward / len(image_paths)


def get_precision_recall(result: np.ndarray, mask: np.ndarray, threshold: float = 0.5):
    pred_mask = result > threshold
    tp = np.logical_and(pred_mask, mask).sum()
    fp = np.logical_and(pred_mask, np.logical_not(mask)).sum()
    fn = np.logical_and(np.logical_not(pred_mask), mask).sum()

    return tp / (tp + fp + 0.0001), tp / (tp + fn + 0.0001)


def get_auc_f1(result: np.ndarray, mask: np.ndarray):
    tot_score = 0
    for thresh in np.arange(-1, 1, 0.1):
        # pr, rec = get_precision_recall(result, mask, thresh)
        tot_score += get_f1_score(result > thresh, mask)

    return tot_score / 20


def get_auc_score(result: np.ndarray, mask: np.ndarray):
    fpr, tpr, _ = roc_curve(mask.ravel(), result.ravel())
    return auc(fpr, tpr)


def get_f1_score(result: np.ndarray, mask: np.ndarray):
    pr, rec = get_precision_recall(result.ravel(), mask.ravel())
    return 2 * pr * rec / (pr + rec + 0.0001)


def get_correlation(result: np.ndarray, mask: np.ndarray):
    return np.abs(np.corrcoef(result.flatten(), mask.flatten())[0, 1])


def get_similarity(result: np.ndarray, mask: np.ndarray):
    return np.dot(result.flatten(), mask.flatten()) / (
        np.linalg.norm(result) * np.linalg.norm(mask)
    )


def get_jaccard(pred, target, threshold=0.5):
    pred_thresh = pred > threshold
    return (
        np.logical_and(pred_thresh, target).sum()
        / np.logical_or(pred_thresh, target).sum()
    )



# -----------------------------------------------------------------------------
# default config values designed to train a gpt2 (124M) on OpenWebText
# I/O
out_dir = "out"
eval_interval = 2000
log_interval = 1
eval_iters = 200
eval_only = False  # if True, script exits right after the first eval
always_save_checkpoint = True  # if True, always save a checkpoint after each eval
init_from = "scratch"  # 'scratch' or 'resume' or 'gpt2*'
# wandb logging
wandb_log = False  # disabled by default
wandb_project = "owt"
wandb_run_name = "gpt2"  # 'run' + str(time.time())
# data
dataset = "openwebtext"
gradient_accumulation_steps = 1  # used to simulate larger batch sizes
batch_size = 12  # if gradient_accumulation_steps > 1, this is the micro-batch size
block_size = 16
# model
n_layer = 4
n_head = 4
n_embd = 128
dropout = 0.0  # for pretraining 0 is good, for finetuning try 0.1+
bias = False  # do we use bias inside LayerNorm and Linear layers?
# adamw optimizer
learning_rate = 1e-4  # max learning rate
max_iters = 2000  # total number of training iterations
weight_decay = 1e-1
beta1 = 0.9
beta2 = 0.95
grad_clip = 1.0  # clip gradients at this value, or disable if == 0.0
# learning rate decay settings
decay_lr = True  # whether to decay the learning rate
warmup_iters = 2000  # how many steps to warm up for
lr_decay_iters = 2000  # should be ~= max_iters per Chinchilla
min_lr = 6e-5  # minimum learning rate, should be ~= learning_rate/10 per Chinchilla
# DDP settings
backend = "nccl"  # 'nccl', 'gloo', etc.
# system
device = "cuda"  # examples: 'cpu', 'cuda', 'cuda:0', 'cuda:1' etc., or try 'mps' on macbooks
dtype = "bfloat16"  # 'float32', 'bfloat16', or 'float16', the latter will auto implement a GradScaler
compile = True  # use PyTorch 2.0 to compile the model to be faster



import os
import random
import torch
import numpy as np

np.seterr(all="raise")


def set_seed(seed: int = 42) -> None:
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    # When running on the CuDNN backend, two further options must be set
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # Set a fixed value for the hash seed
    os.environ["PYTHONHASHSEED"] = str(seed)
    print(f"Random seed set as {seed}")


def standardize(image: np.ndarray, max_z: float = 3, min_max: tuple = (0, 1)) -> np.ndarray:
    img_mean = image.mean(axis=1)[:, None]
    img_std = image.std(axis=1)[:, None]
    image = (image - img_mean) / (img_std + 0.0000001)

    image = (np.clip(image, -max_z, max_z) + max_z) / (2 * max_z)
    image = image * (min_max[1] - min_max[0]) + min_max[0]
    return image


def min_max_normalize(image: np.ndarray) -> np.ndarray:
    img_min = image.min(axis=1)[:, None]
    img_max = image.max(axis=1)[:, None]
    return (image - img_min) / (img_max - img_min)
